"""Base class for monthly collections reports for the City of Philadelphia."""

import calendar
from dataclasses import dataclass
from typing import ClassVar

import pdfplumber

from ... import DATA_DIR
from ...etl import ETLPipeline
from ...utils.misc import fiscal_from_calendar_year
from ...utils.transformations import *


def get_column_names(month: int, calendar_year: int) -> List[str]:
    """Columns for monthly collection reports."""

    # Get the FY from the calendar year
    fiscal_year = fiscal_from_calendar_year(month, calendar_year)

    # Get the month name
    month_name = calendar.month_abbr[month].lower()

    # Fiscal year tsags
    this_year = f"fy{str(fiscal_year)[2:]}"
    last_year = f"fy{str(fiscal_year-1)[2:]}"

    return [
        f"{last_year}_actual",
        f"{this_year}_budgeted",
        f"{month_name}_{this_year}",
        f"{month_name}_{last_year}",
        f"{this_year}_ytd",
        f"{last_year}_ytd",
        "net_change",
        "budget_requirement",
        "pct_budgeted",
    ]


@dataclass
class MonthlyCollectionsReport(ETLPipeline):
    """
    Base class for extracting data from the City of Philadelphia's
    monthly collections PDF reports.

    Parameters
    ----------
    month :
        the calendar month number (starting at 1)
    year :
        the calendar year
    """

    report_type: ClassVar[str]
    month: int
    year: int

    def __post_init__(self):
        """Set up necessary variables."""

        # The PDF path
        dirname = self.get_data_directory("raw")
        self.path = dirname / f"{self.year}_{self.month:02d}.pdf"

        # Make sure this path exists
        if not self.path.exists():
            raise FileNotFoundError(
                f"No PDF available for month '{self.month}' and year '{self.year}' at '{self.path}'"
            )

        # Get the number of pages
        with pdfplumber.open(self.path) as pdf:
            self.num_pages = len(pdf.pages)

        # Month name
        self.month_name = calendar.month_abbr[self.month].lower()

    @classmethod
    def get_data_directory(cls, kind: str) -> str:
        """Internal function to get the file path.

        Parameters
        ----------
        kind : {'raw', 'processed'}
            type of data to load
        """
        assert kind in ["raw", "processed"]

        return DATA_DIR / kind / "collections" / "monthly" / f"{cls.report_type}"

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Transform the raw parsing data into a clean data frame."""

        # Apply each of the transformations
        data = (
            data.pipe(remove_footnotes)
            .pipe(fix_duplicated_chars)
            .pipe(remove_spaces)
            .pipe(fix_duplicate_parens)
            .pipe(fix_percentages)
            .pipe(replace_missing_cells)
            .pipe(remove_extra_letters, col_num=1)
            .pipe(convert_to_floats, usecols=data.columns[1:])
            .pipe(remove_missing_rows, usecols=data.columns[1:])
        ).reset_index(drop=True)

        return data



from .by_industry.sales import SalesCollectionsByIndustry
from .by_industry.wage import WageCollectionsByIndustry
from .monthly.city_nontax import CityNonTaxCollections
from .monthly.city_other_govts import CityOtherGovtsCollections
from .monthly.city_tax import CityTaxCollections
from .monthly.school import SchoolTaxCollections



import itertools
import re
from dataclasses import dataclass
from operator import attrgetter
from typing import Dict, Iterator, List, Optional, Type, TypeVar

import desert
import marshmallow
import numpy as np
import pandas as pd
import pdfplumber
from intervaltree import IntervalTree

# Create a generic variable that can be 'Parent', or any subclass.
Word_T = TypeVar("Word_T", bound="Word")


def extract_words(
    pg: pdfplumber.page.Page,
    keep_blank_chars: bool = False,
    x_tolerance: int = 2,
    y_tolerance: int = 2,
):
    words = []

    for word_dict in pg.extract_words(
        keep_blank_chars=keep_blank_chars,
        x_tolerance=x_tolerance,
        y_tolerance=y_tolerance,
    ):

        # Convert to a Word
        word = Word.from_dict(word_dict)

        # Clean up text
        word.text = word.text.replace("\xa0", " ").strip()

        # Save it
        if word.text:
            words.append(word)

    # Sort the words top to bottom and left to right
    return sorted(words, key=attrgetter("top", "x0"), reverse=False)


@dataclass
class Word:
    """
    A word in the PDF with associated text and bounding box.

    Parameters
    ----------
    x0 :
        the starting horizontal coordinate
    x1 :
        the ending horizontal coordinate
    bottom :
        the bottom vertical coordinate
    top :
        the top vertical coordinate
    text :
        the associated text
    """

    x0: float
    x1: float
    top: float
    bottom: float
    text: str

    @property
    def x(self) -> float:
        """Alias for `x0`."""
        return self.x0

    @property
    def y(self) -> float:
        """Alias for `tops`."""
        return self.top

    @classmethod
    def from_dict(cls: Type[Word_T], data: dict) -> Word_T:
        """
        Return a new class instance from a dictionary
        representation.

        Parameters
        ----------
        data :
            The dictionary representation of the class.
        """
        schema = desert.schema(cls, meta={"unknown": marshmallow.EXCLUDE})
        return schema.load(data)


def groupby(words: List[Word], key: str, sort: bool = False) -> Iterator:
    """Group words by the specified attribute, optionally sorting."""
    if sort:
        words = sorted(words, key=attrgetter(key))
    return itertools.groupby(words, attrgetter(key))


def fuzzy_groupby(
    words: List[Word], lower_tol: int = 10, upper_tol: int = 10, key="y"
) -> Dict[float, List[Word]]:
    """Group words into lines, with a specified tolerance."""

    tree = IntervalTree()
    for i in range(len(words)):
        y = getattr(words[i], key)
        tree[y - lower_tol : y + upper_tol] = words[i]  # type: ignore

    result: Dict[float, List[Word]] = {}
    for y in sorted(np.unique([getattr(w, key) for w in words])):
        objs = [iv.data for iv in tree[y]]
        values = sorted(objs, key=attrgetter("x"))

        if values not in result.values():
            result[y] = values

    return result


def create_data_table(headers, columns, match_tol=20) -> pd.DataFrame:
    """Based on headers and column data, create the data table."""

    # Store the bottom y values of all of the row headers
    header_tops = np.array([h.top for h in headers])

    # Set up the grid: nrows by ncols
    nrows = len(headers)
    ncols = len(columns) + 1

    # Initialize the grid
    grid = np.empty((nrows, ncols), dtype=object)
    grid[:, :] = ""  # Default value

    # Add in the headers
    grid[:, 0] = [h.text for h in headers]

    # Loop over each column
    for col_num, xval in enumerate(columns):

        col = columns[xval]
        word_tops = np.array([w.top for w in col])

        # Find closest row header
        for row_num, h in enumerate(headers):

            # Find closest word ot this row heasder
            word_diff = np.abs(word_tops - h.top)
            word_diff[word_diff > match_tol] = np.nan

            # Make sure the row header is vertically close enough
            if np.isnan(word_diff).sum() < len(word_diff):

                # Get the matching word for this row header
                notnull = ~np.isnan(word_diff)
                order = np.argsort(word_diff[notnull])
                for word_index in np.where(notnull)[0][order]:
                    word = col[word_index]

                    # IMPORTANT: make sure this is the closest row header
                    # Sometimes words will match to more than one header
                    header_diff = np.abs(header_tops - word.top)
                    header_index = np.argmin(header_diff)
                    closest_header = headers[header_index]
                    if closest_header == h:
                        grid[row_num, col_num + 1] = col[word_index].text
                        break

    return pd.DataFrame(grid)


def remove_all_alpha_columns(columns):
    """Remove orphan columns that are all words and no numbers."""
    for key in sorted(columns):
        value = columns[key]
        if all([w.text.replace(" ", "").isalpha() for w in value]):
            del columns[key]
    return columns


def remove_close_columns(columns, min_col_sep=24):
    """Remove columns that are closer than a specified distance."""

    locations = sorted(columns)
    for i in range(len(locations) - 1):
        this_loc = locations[i]
        next_loc = locations[i + 1]

        if this_loc not in columns:
            continue

        if next_loc - this_loc < min_col_sep:
            if len(columns[next_loc]) > len(columns[this_loc]):
                columns[next_loc] += columns[this_loc]
                del columns[this_loc]
            else:
                columns[this_loc] += columns[next_loc]
                del columns[next_loc]

    return columns


def remove_orphan_columns(columns):
    """Remove any columns that are full subsets of another column."""
    remove = []
    for k in columns:
        values = columns[k]
        if any([all(val in columns[j] for val in values) for j in columns if j != k]):
            remove.append(k)

    return {k: columns[k] for k in columns if k not in remove}


def remove_hyphens(df):

    df = df.replace("-", "")
    bad_columns = (df == "").all(axis=0)
    if bad_columns.sum():
        df = df[df.columns[~bad_columns]]
        df.columns = list(range(0, len(df.columns)))

    return df


def words_to_table(
    words: List[Word],
    text_tolerance_y: int = 5,
    text_tolerance_x: int = 3,
    column_tolerance: int = 20,
    min_col_sep: int = 24,
    header_column_overlap: int = 10,
) -> pd.DataFrame:
    """Combine words into a table, returning a pandas DataFrame"""

    # Group into words
    rows = fuzzy_groupby(
        words, key="bottom", lower_tol=text_tolerance_y, upper_tol=text_tolerance_y
    )

    # Split headers (first column) from rest
    words = []
    headers = []

    # Loop over phrases
    footnote_cutoff = None
    for k in rows:
        row = rows[k]

        # Group words into phrases
        for i in reversed(range(1, len(row))):
            this_word = row[i]
            prev_word = row[i - 1]

            if this_word.x0 - prev_word.x1 < text_tolerance_x:
                row[i - 1].text += " " + row[i].text
                row[i - 1].x1 = row[i].x1
                del row[i]

        # Skip footnote rows
        first = row[0].text
        if first.startswith("*"):
            continue

        # Split headers from numbers
        if first[0].strip().isalpha():

            # Keep the header
            headers.append(row[0])

            # Keep any non-header numbers
            words += [
                w
                for w in row
                if re.match(
                    "\(?(\d+(?:\.\d+)?%|\d{1,3}(?:,\d{3})*|-|N/A)\)?",
                    w.text.replace(" ", ""),
                )
            ]
        else:
            words += [w for w in row]

    # Group into columns and sort it
    columns = fuzzy_groupby(
        words, lower_tol=column_tolerance, upper_tol=column_tolerance, key="x1"
    )
    columns = {key: sorted(columns[key], key=attrgetter("top")) for key in columns}

    # Remove any columns that are full subsets of another column
    # This can happen when fuzzy matching to do the columns
    columns = remove_orphan_columns(columns)

    # Remove any orphan columns that are too close together
    # Sometimes columns will end up split when they should be merged
    columns = remove_close_columns(columns, min_col_sep=min_col_sep)

    # Text in header column are sometimes split, leaving parts of words
    # in an orphan column — this checks if columns are all alpha
    # and removes them
    columns = remove_all_alpha_columns(columns)

    # Create the table
    table = create_data_table(headers, columns, match_tol=header_column_overlap)

    return remove_hyphens(table)



"""Command-line interface."""

import click
from loguru import logger

from .etl import DATA_DIR, collections


@click.command()
@click.version_option()
@click.argument(
    "kind", type=click.Choice(["city-tax", "city-nontax", "city-other-govts", "wage"])
)
@click.option("--month", type=int)
@click.option("--year", type=int)
@click.option("--dry-run", is_flag=True)
def monthly_collections_etl(kind, month=None, year=None, dry_run=False) -> None:
    """Run the ETL pipeline for monthly collections"""

    # Get the ETL class
    ETL = {
        "wage": collections.WageCollectionsByIndustry,
        "city-tax": collections.CityTaxCollections,
        "city-nontax": collections.CityNonTaxCollections,
        "city-other-govts": collections.CityOtherGovtsCollections,
    }
    cls = ETL[kind]

    # Log
    logger.info(f"Processing ETL for '{cls.__name__}'")

    # If month is provided, we need the year too
    if month is not None:
        if year is None:
            raise ValueError("Year is required if month is provided")

    # Do a single month and year
    if month is not None:

        # Log
        logger.info(f"Processing year='{year}' and month='{month}'")

        # Do the ETL
        if not dry_run:
            report = cls(year=year, month=month)
            report.extract_transform_load()
    else:

        # Get the directory of raw files
        dirname = cls.get_data_directory("raw")

        # Glob the PDF files
        if year is not None:
            files = dirname.glob(f"{year}*.pdf")
        else:
            files = dirname.glob("*.pdf")

        # Do all the files
        for f in sorted(files):

            # Get month and year
            year, month = map(int, f.stem.split("_"))
            logger.info(f"Processing year='{year}' and month='{month}'")

            # ETL
            if not dry_run:
                report = cls(year=year, month=month)
                report.extract_transform_load()



def fiscal_from_calendar_year(month_num, calendar_year):
    """Return the fiscal year for the input calendar year."""

    return calendar_year if month_num < 7 else calendar_year + 1


def rename_tax_rows(df, index, tax_names, suffixes=["current", "prior", "total"]):
    """Internal function that loops over consecutive rows and adds the name."""

    for tax_name in tax_names:
        for offset in [0, 1, 2]:
            suffix = suffixes[offset]
            df.loc[index + offset, 0] = f"{tax_name}_{suffix}"
        index = index + 3

    return index



from dataclasses import dataclass
from operator import attrgetter

import pdfplumber

from ... import DATA_DIR
from ...etl import ETLPipeline
from ...utils.pdf import extract_words, fuzzy_groupby
from ...utils.transformations import *

INDUSTRIES = [
    "All Other Sectors",
    "Appliance, other electronics, retail",
    "Car and truck rental",
    "Computer and software stores, retail",
    "Construction",
    "Convenience stores, retail",
    "Department stores, retail",
    "Furniture stores retail",
    "Home centers, retail",
    "Hotels",
    "Liquor and beer stores, retail",
    "Manufacturing",
    "Motor Vehicle Sales Tax",
    "Office supplies stores, retail",
    "Other retail",
    "Pharmacies, retail",
    "Public Utilities",
    "Rentals except car and truck rentals",
    "Repair services",
    "Restaurants, bars, concessionaires and caterers",
    "Services other than repair services",
    "Subtotal",
    "Supermarkets, retail",
    "Telecommunications",
    "Total Retail",
    "Unclassified",
    "Wholesale",
]


@dataclass
class SalesCollectionsByIndustry(ETLPipeline):
    """
    Fiscal year sales collections by industry.

    Parameters
    ----------
    fiscal_year :
        the fiscal year; data is annual
    """

    fiscal_year: int

    def __post_init__(self):
        """Set up necessary variables."""

        # The PDF path
        fy_tag = str(self.fiscal_year)[-2:]
        self.path = self.get_data_directory("raw") / f"FY{fy_tag}.pdf"

        # Make sure this path exists
        if not self.path.exists():
            raise FileNotFoundError(
                f"No PDF available for fiscal year '{self.fiscal_year}'"
            )

        # Which file format?
        self.legacy = self.fiscal_year < 2017

    @classmethod
    def get_data_directory(cls, kind: str) -> str:
        """Internal function to get the file path.

        Parameters
        ----------
        kind : {'raw', 'processed'}
            type of data to load
        """
        assert kind in ["raw", "processed"]

        return DATA_DIR / kind / "collections" / "by-industry" / "sales"

    def extract(self) -> pd.DataFrame:
        """Extract the data from the first PDF page."""

        # Open the PDF document
        with pdfplumber.open(self.path) as pdf:

            # Only need first page
            pg = pdf.pages[0]

            # Determine crop areas
            all_words = extract_words(
                pg, keep_blank_chars=True, x_tolerance=2, y_tolerance=1
            )

            ## TOP LEFT
            top_left = [
                w
                for w in all_words
                if w.text.strip().lower().startswith("construction")
            ]
            top_left = min(top_left, key=lambda w: w.x0)

            ## BOTTOM LEFT
            bottom_left = [
                w for w in all_words if w.text.strip().lower().startswith("motor")
            ]
            bottom_left = min(bottom_left, key=lambda w: w.x0)

            # Crop the main part of the document and extract the words
            cropped = pg.crop(
                [pg.bbox[0], top_left.top, pg.bbox[2], bottom_left.bottom + 3]
            )

            # Table strategy based on format
            if self.legacy:
                horizontal_strategy = "lines"
            else:
                horizontal_strategy = "text"

            return pd.DataFrame(
                cropped.extract_table(
                    {
                        "vertical_strategy": "lines",
                        "horizontal_strategy": horizontal_strategy,
                    }
                ),
            )

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Transform the raw parsing data into a clean data frame."""

        # Apply set of base transformations first
        data = (
            data.pipe(remove_spaces)
            .pipe(fix_percentages)
            .pipe(strip_dollar_signs)
            .pipe(replace_missing_cells)
            .pipe(convert_to_floats, usecols=data.columns[1:])
        ).reset_index(drop=True)

        # Check length
        assert len(data) == 27

        if not self.legacy:
            assert len(data.columns) == 12
            data = data[[0, 1, 2, 3]]
        else:
            assert len(data.columns) == 4

        # Set the columns
        data.columns = ["industry", "number_entities", "total", "percent_of_total"]

        # Sort by first column
        data = data.sort_values("industry")

        # Assign uniform industries
        data["industry"] = INDUSTRIES
        data["parent_industry"] = np.select(
            [
                data["industry"].isin(
                    [
                        "Furniture stores retail",
                        "Appliance, other electronics, retail",
                        "Computer and software stores, retail",
                        "Home centers, retail",
                        "Supermarkets, retail",
                        "Convenience stores, retail",
                        "Liquor and beer stores, retail",
                        "Pharmacies, retail",
                        "Department stores, retail",
                        "Office supplies stores, retail",
                        "Other retail",
                    ]
                ),
            ],
            [
                "Total Retail",
            ],
            default="",
        )
        data["parent_industry"] = data["parent_industry"].replace("", np.nan)

        return data.sort_index()

    def validate(self, data):
        """Validate the input data."""

        # Sum up
        main_industries = data.query(
            "parent_industry.isnull() and industry != 'Subtotal' and industry != 'Motor Vehicle Sales Tax'"
        )
        subtotal1 = main_industries["total"].sum()
        subtotal2 = data.query("industry == 'Subtotal'")["total"].squeeze()
        diff = subtotal1 - subtotal2
        assert diff < 5

        # Sub industries
        subsectors = data.query("parent_industry.notnull()")
        totals = subsectors.groupby("parent_industry")["total"].sum()

        # Compare to total
        for industry in totals.index:
            total1 = totals.loc[industry]
            total2 = data.loc[data["industry"] == industry]["total"].squeeze()
            diff = total1 - total2
            assert diff < 5

        return True

    def load(self, data) -> None:
        """Load the data."""

        # Get the path
        fy_tag = str(self.fiscal_year)[-2:]
        path = self.get_data_directory("processed") / f"FY{fy_tag}.csv"

        # Load
        super()._load_csv_data(path)



from ...utils.misc import rename_tax_rows
from ...utils.transformations import remove_missing_rows
from .city import CityCollectionsReport
from .core import get_column_names


class CityTaxCollections(CityCollectionsReport):
    """Monthly City Tax Collections Report"""

    def transform(self, data):
        """Transform."""

        # Call base transform
        data = super().transform(data)

        # Split out just tax part
        stop = data.index[data[0].str.contains("TOTAL TAX REVENUE") == True]
        assert len(stop) == 1
        stop = stop[0]

        # Trim
        tax = data.loc[:stop].copy().dropna(how="all", axis=1)

        # Remove Data Warehouse
        sel = tax[0].str.contains("DATA WAREHOUSE")
        if sel.sum():
            tax = tax.loc[~sel]
            tax = tax.reset_index(drop=True)

        # Determine columns for the report
        columns = get_column_names(self.month, self.year)
        columns = ["name"] + columns[-7:]
        tax = tax[[0] + list(tax.columns[-7:])]

        # Remove any empty rows
        tax = tax.pipe(remove_missing_rows, usecols=tax.columns[1:]).reset_index(
            drop=True
        )

        # Check dimensions
        assert len(columns) == len(tax.columns)
        assert len(tax) in [39, 40, 42], len(tax)

        # Real estate + wage
        index = rename_tax_rows(
            tax,
            0,
            [
                "real_estate",
                "wage_city",
                "wage_pica",
            ],
        )
        tax.loc[index, 0] = "wage_total"
        index += 1

        # Earnings
        index = rename_tax_rows(
            tax,
            index,
            [
                "earnings_city",
                "earnings_pica",
            ],
        )
        tax.loc[index, 0] = "earnings_total"
        index += 1

        # Net Profits
        index = rename_tax_rows(
            tax,
            index,
            ["net_profits_city", "net_profits_pica"],
        )
        tax.loc[index, 0] = "net_profits_total"
        index += 1

        # Combo of wage, earnings, and NPT
        for i, suffix in enumerate(["", "pica", "city"]):

            if suffix:
                suffix = f"{suffix}_total"
            else:
                suffix = "total"
            tax.loc[index + i, 0] = f"wage_earnings_net_profits_{suffix}"
        index += 3

        # Add birt
        index = rename_tax_rows(
            tax,
            index,
            ["birt"],
        )

        # Other taxes
        other_taxes = [
            name + "_total"
            for name in [
                "sales",
                "amusement",
                "tobacco",
                "parking",
                "valet",
                "real_estate_transfer",
                "outdoor_ads",
            ]
        ]
        tax.loc[index : index + len(other_taxes) - 1, 0] = other_taxes
        index = index + len(other_taxes)

        # Handle remaining rows
        if len(tax) == 42:
            remaining = [
                "soda_current",
                "soda_prior",
                "soda_total",
                "other_taxes_total",
                "all_taxes_total",
            ]
        elif len(tax) == 40:
            remaining = ["soda_total", "other_taxes_total", "all_taxes_total"]
        elif len(tax):
            remaining = ["other_taxes_total", "all_taxes_total"]

        n = len(remaining)
        tax.loc[index : index + n - 1, 0] = remaining

        # Rename the columns
        tax.columns = columns

        # Split out current/prior/total into its own column
        tax["kind"] = tax["name"].apply(lambda x: x.split("_")[-1])
        tax["name"] = tax["name"].apply(lambda x: "_".join(x.split("_")[:-1]))

        return tax

    def load(self, data):
        """Load the data into storage."""

        # Get the processed data path
        dirname = self.get_data_directory("processed")
        path = dirname / f"{self.year}-{self.month:02d}-tax.csv"

        # Load
        super()._load_csv_data(data, path)

    def validate(self, data):
        """Validate the input data."""

        taxes = [
            "real_estate",
            "wage_city",
            "earnings_city",
            "net_profits_city",
            "birt",
            "sales",
            "amusement",
            "tobacco",
            "parking",
            "valet",
            "real_estate_transfer",
            "outdoor_ads",
            "soda",
            "other_taxes",
        ]
        t = data.query("kind == 'total' and name in @taxes")
        t = t.filter(regex=f"^{self.month_name}", axis=1)

        for col in t.columns:
            all_taxes = data.query("name == 'all_taxes'")[col].squeeze()
            diff = t[col].sum() - all_taxes
            assert diff < 5

        return True



import pandas as pd
import pdfplumber

from ... import DATA_DIR
from ...utils.misc import rename_tax_rows
from ...utils.pdf import extract_words, words_to_table
from .core import MonthlyCollectionsReport, get_column_names


class SchoolTaxCollections(MonthlyCollectionsReport):
    """
    Monthly School District Collections Report.

    Parameters
    ----------
    month :
        the calendar month number (starting at 1)
    year :
        the calendar year
    """

    report_type = "school"

    @property
    def legacy(self) -> bool:
        """Whether the format is the legacy or current version."""
        return self.num_pages > 1

    def extract(self) -> pd.DataFrame:
        """Internal function to parse the contents of a legacy PDF page."""

        # Open the PDF document
        with pdfplumber.open(self.path) as pdf:

            # Loop over each page
            out = []
            for pg in pdf.pages:

                # Extract the words
                words = extract_words(
                    pg, keep_blank_chars=False, x_tolerance=1, y_tolerance=1
                )

                # Group the words into a table
                data = words_to_table(
                    words,
                    text_tolerance_y=5,
                    text_tolerance_x=5,
                    column_tolerance=20,
                    min_col_sep=24,
                    header_column_overlap=10,
                )

                # Skip the header (first five rows)
                data = data.iloc[6:]
                assert "REAL ESTATE" in data.iloc[0][0]

                # # Remove first row of header if we need to
                # for phrase in ["prelim", "final", "budget"]:
                #     sel = data[0].str.lower().str.startswith(phrase)
                #     data = data.loc[~sel]

                # # Remove empty columns
                # data = remove_empty_columns(data, use_nan=False)

                # Check number of columns
                if len(out):
                    if len(data.columns) != len(out[-1].columns):
                        raise ValueError("Column mismatch when parsing multiple pages")

                # Save it
                out.append(data)

            # Return concatenation
            return pd.concat(out, axis=0, ignore_index=True)

    def transform(self, data):
        """Transform the raw parsing data into a clean data frame."""

        # Call base transform
        data = super().transform(data)

        # Determine columns for the report
        columns = get_column_names(self.month, self.year)

        ncols = len(data.columns)
        assert ncols in [11, 12, 14]

        if ncols == 14:
            data = data.drop(labels=[7, 8, 9, 10], axis=1)
        else:
            data = data.drop(labels=[data.columns[-6]], axis=1)

        # Set the columns
        columns = ["name"] + columns[-7:]
        data = data[[0] + list(data.columns[-7:])]

        assert len(columns) == len(data.columns)
        assert len(data) in [14, 15]

        # Do current/prior/total
        if len(data) == 14:
            index = rename_tax_rows(
                data,
                0,
                ["real_estate", "school_income", "use_and_occupancy", "liquor"],
            )
        else:
            index = rename_tax_rows(
                data,
                0,
                ["real_estate"],  # , "school_income", "use_and_occupancy", "liquor"],
            )

            if "PAYMENT" in data.loc[index, 0]:
                data.loc[index, 0] = "pilots_total"
                index += 1

            index = rename_tax_rows(
                data,
                index,
                ["school_income", "use_and_occupancy", "liquor"],
            )

            if "PAYMENT" in data.loc[index, 0]:
                data.loc[index, 0] = "pilots_total"
                index += 1

        # Other non-tax
        data.loc[index, 0] = "other_nontax_total"
        index += 1

        # Total
        data.loc[index, 0] = "total_revenue_total"
        index += 1

        # Set the columns
        data.columns = columns

        # Split out current/prior/total into its own column
        data["kind"] = data["name"].apply(lambda x: x.split("_")[-1])
        data["name"] = data["name"].apply(lambda x: "_".join(x.split("_")[:-1]))

        return data

    def validate(self, data):
        """Validate the input data."""

        # Sum up
        t = data.query("kind == 'total' and name != 'total_revenue'")
        t = t.filter(regex=f"^{self.month_name}", axis=1)

        # Compare to total
        for col in t.columns:
            total_revenue = data.query("name == 'total_revenue'")[col].squeeze()
            diff = t[col].sum() - total_revenue
            assert diff < 5

        return True

    def load(self, data) -> None:
        """Load the data."""

        # Get the path
        dirname = self.get_data_directory("processed")
        path = dirname / f"{self.year}-{self.month:02d}-tax.csv"

        # Load
        super()._load_csv_data(data, path)



import calendar
from dataclasses import dataclass
from operator import attrgetter

import pdfplumber

from ... import DATA_DIR
from ...etl import ETLPipeline
from ...utils.pdf import extract_words, fuzzy_groupby
from ...utils.transformations import *

INDUSTRIES = [
    "Construction",
    "Manufacturing",
    "Chemicals, Petroleum Refining",
    "Pharmaceuticals",
    "Transportation Equipment",
    "Food & Beverage Products",
    "Machinery, Electronic, and Other Electric Equipment",
    "Metal Manufacturing",
    "Miscellaneous Manufacturing",
    "Public Utilities",
    "Transportation and Warehousing",
    "Telecommunication",
    "Publishing, Broadcasting, and Other Information",
    "Wholesale Trade",
    "Retail Trade",
    "Banking & Credit Unions",
    "Securities / Financial Investments",
    "Insurance",
    "Real Estate, Rental and Leasing",
    "Health and Social Services",
    "Hospitals",
    "Doctors, Dentists, and Other Health Practitioners",
    "Outpatient Care Centers and Other Health Services",
    "Nursing & Personal Care Facilities",
    "Social Services",
    "Education",
    "College and Universities",
    "Elementary, Secondary Schools",
    "Other Educational Services",
    "Professional Services",
    "Legal Services",
    "Management Consulting",
    "Engineering & Architectural Services",
    "Computer",
    "Accounting, Auditing, Bookkeeping",
    "Advertising and Other Professional Services",
    "Hotels",
    "Restaurants",
    "Sport Teams",
    "Arts, Entertainment, and Other Recreation",
    "Other Sectors",
    "Membership Organizations",
    "Employment/Outsourcing Agencies",
    "Security and Investigation Services",
    "Services to Buildings",
    "Miscellaneous Sectors",
    "Government",
    "State Government (PA)",
    "City, School District, Local Quasi Govt.",
    "Federal Government",
    "Other Governments",
    "Unclassified Accounts",
]


@dataclass
class WageCollectionsByIndustry(ETLPipeline):
    """
    Monthly wage collections by industry.

    Parameters
    ----------
    month :
        the calendar month number (starting at 1)
    year :
        the calendar year
    """

    month: int
    year: int

    def __post_init__(self):
        """Set up necessary variables."""

        # The PDF path
        self.path = self.get_data_directory("raw") / f"{self.year}_{self.month:02d}.pdf"

        # Make sure this path exists
        if not self.path.exists():
            raise FileNotFoundError(
                f"No PDF available for month '{self.month}' and year '{self.year}'"
            )

        # Month name
        self.month_name = calendar.month_abbr[self.month].lower()

    @classmethod
    def get_data_directory(cls, kind: str) -> str:
        """Internal function to get the file path.

        Parameters
        ----------
        kind : {'raw', 'processed'}
            type of data to load
        """
        assert kind in ["raw", "processed"]
        return DATA_DIR / kind / "collections" / "by-industry" / "wage"

    def extract(self) -> pd.DataFrame:
        """Extract the data from the first PDF page."""

        # Open the PDF document
        with pdfplumber.open(self.path) as pdf:

            # Only need first page
            pg = pdf.pages[0]

            # Determine crop areas
            all_words = extract_words(
                pg, keep_blank_chars=True, x_tolerance=2, y_tolerance=1
            )

            ## TOP LEFT
            top_left = [
                w
                for w in all_words
                if w.text.strip().lower().startswith("construction")
            ]
            top_left = min(top_left, key=lambda w: w.x0)

            ## BOTTOM LEFT
            bottom_left = [
                w
                for w in all_words
                if w.text.strip().lower().startswith("unclassified")
            ]
            bottom_left = min(bottom_left, key=lambda w: w.x0)

            # Crop the main part of the document and extract the words
            cropped = pg.crop(
                [top_left.x0, top_left.top, pg.bbox[2], bottom_left.bottom]
            )
            words = extract_words(
                cropped, keep_blank_chars=True, x_tolerance=2, y_tolerance=1
            )

            # Group into rows
            d = []
            for k, v in fuzzy_groupby(
                words, key="bottom", lower_tol=3, upper_tol=3
            ).items():
                d.append([w.text for w in sorted(v, key=attrgetter("x0"))])

            return pd.DataFrame(d)

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Transform the raw parsing data into a clean data frame."""

        # Apply set of base transformations first
        data = (
            data.pipe(remove_spaces)
            .pipe(fix_percentages)
            .pipe(replace_missing_cells)
            .pipe(convert_to_floats, usecols=data.columns[1:])
        ).reset_index(drop=True)

        # Check length
        assert len(data) == 52
        assert len(data.columns) == 10

        columns = ["industry"]

        # Monthly collections over time
        for i in [0, 1, 2, 3]:
            columns.append(f"{self.month_name}_{self.year-i}")

        # Growth rates
        for i in [0, 1, 2]:
            columns.append(f"growth_yoy_{self.year-i}")
        columns += ["growth_3yr", "net_change"]

        # Set the columns
        data.columns = columns

        # Assign uniform industries
        data["industry"] = INDUSTRIES
        data["parent_industry"] = np.select(
            [
                data["industry"].isin(
                    [
                        "Chemicals, Petroleum Refining",
                        "Pharmaceuticals",
                        "Transportation Equipment",
                        "Food & Beverage Products",
                        "Machinery, Electronic, and Other Electric Equipment",
                        "Metal Manufacturing",
                        "Miscellaneous Manufacturing",
                    ]
                ),
                data["industry"].isin(
                    [
                        "Hospitals",
                        "Doctors, Dentists, and Other Health Practitioners",
                        "Outpatient Care Centers and Other Health Services",
                        "Nursing & Personal Care Facilities",
                        "Social Services",
                    ]
                ),
                data["industry"].isin(
                    [
                        "College and Universities",
                        "Elementary, Secondary Schools",
                        "Other Educational Services",
                    ]
                ),
                data["industry"].isin(
                    [
                        "Legal Services",
                        "Management Consulting",
                        "Engineering & Architectural Services",
                        "Computer",
                        "Accounting, Auditing, Bookkeeping",
                        "Advertising and Other Professional Services",
                    ]
                ),
                data["industry"].isin(
                    [
                        "Membership Organizations",
                        "Employment/Outsourcing Agencies",
                        "Security and Investigation Services",
                        "Services to Buildings",
                        "Miscellaneous Sectors",
                    ]
                ),
                data["industry"].isin(
                    [
                        "State Government (PA)",
                        "City, School District, Local Quasi Govt.",
                        "Federal Government",
                        "Other Governments",
                    ]
                ),
            ],
            [
                "Manufacturing",
                "Health and Social Services",
                "Education",
                "Professional Services",
                "Other Sectors",
                "Government",
            ],
            default="",
        )
        data["parent_industry"] = data["parent_industry"].replace("", np.nan)

        return data

    def validate(self, data):
        """Validate the input data."""

        cols = [f"{self.month_name}_{self.year-i}" for i in [0, 1, 2, 3]]

        # Sum up
        subsectors = data.query("parent_industry.notnull()")
        totals = subsectors.groupby("parent_industry")[cols].sum()

        # Compare to total
        for col in cols:
            for industry in totals.index:
                total1 = totals.loc[industry, col]
                total2 = data.loc[data["industry"] == industry][col].squeeze()
                diff = total1 - total2
                assert diff < 5

        return True

    def load(self, data) -> None:
        """Load the data."""

        # Path to save data to
        dirname = self.get_data_directory("processed")
        path = dirname / f"{self.year}-{self.month:02d}.csv"

        # Load
        super()._load_csv_data(data, path)



import os
import time
from collections import defaultdict
from dataclasses import dataclass
from typing import ClassVar

import boto3
import pandas as pd
from dotenv import load_dotenv

from .utils.pdf import Word


def map_blocks(blocks, block_type):
    return {block["Id"]: block for block in blocks if block["BlockType"] == block_type}


def textract_to_words(response):
    """Return words from AWS Textract response."""

    # Extract block types
    blocks = response["Blocks"]
    words = map_blocks(blocks, "WORD")

    out = defaultdict(list)
    for k in words:
        w = words[k]
        bbox = w["Geometry"]["BoundingBox"]
        out[w["Page"]].append(
            Word(
                x0=bbox["Left"],
                x1=bbox["Left"] + bbox["Width"],
                top=bbox["Top"],
                bottom=bbox["Top"] + bbox["Height"],
                text=w["Text"],
            )
        )

    for pg_num in sorted(out):
        yield out[pg_num]


def textract_to_table(response):
    """Yield tables from AWS Textract response as pandas DataFrames."""

    # Extract block types
    blocks = response["Blocks"]
    tables = map_blocks(blocks, "TABLE")
    cells = map_blocks(blocks, "CELL")
    words = map_blocks(blocks, "WORD")
    selections = map_blocks(blocks, "SELECTION_ELEMENT")

    def get_children_ids(block):
        for rels in block.get("Relationships", []):
            if rels["Type"] == "CHILD":
                yield from rels["Ids"]

    # Look over each of the tables
    for table in tables.values():

        # Determine all the cells that belong to this table
        table_cells = [cells[cell_id] for cell_id in get_children_ids(table)]

        # Determine the table's number of rows and columns
        n_rows = max(cell["RowIndex"] for cell in table_cells)
        n_cols = max(cell["ColumnIndex"] for cell in table_cells)
        content = [[None for _ in range(n_cols)] for _ in range(n_rows)]

        # Fill in each cell
        for cell in table_cells:
            cell_contents = [
                words[child_id]["Text"]
                if child_id in words
                else selections[child_id]["SelectionStatus"]
                for child_id in get_children_ids(cell)
            ]
            i = cell["RowIndex"] - 1
            j = cell["ColumnIndex"] - 1
            content[i][j] = " ".join(cell_contents)

        yield table["Page"], pd.DataFrame(content)


@dataclass
class AWSTextract:
    """
    Interface for AWS Textract to extract tables from PDFs.

    Notes
    -----
    - Data is uploaded to AWS s3 before passing to Textract.
    - The 'AWS_ACCESS_KEY' and 'AWS_SECRET_KEY' keys should be
    specified as environment files.

    Parameters
    ----------
    path :
        the local path to the PDF to extract tables from
    """

    path: str
    BUCKET: ClassVar[str] = "phl-budget-data"

    def __post_init__(self):
        """Initialize the AWS clients."""

        # Load env variables
        load_dotenv()

        # Make sure we have the keys
        aws_access_key_id = os.environ.get("AWS_ACCESS_KEY")
        if aws_access_key_id is None:
            raise ValueError("Specify AWS_ACCESS_KEY as environment variable")
        aws_secret_access_key = os.environ.get("AWS_SECRET_KEY")
        if aws_secret_access_key is None:
            raise ValueError("Specify AWS_SECRET_KEY as environment variable")

        # Initialize the aws clients
        self.s3 = boto3.client(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )

        self.textract = boto3.client(
            "textract",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
        )

        # Create the bucket if we need to
        self.bucket = self.s3.create_bucket(Bucket=self.BUCKET)

    def extract(self):
        """Extract data using AWS Textract."""

        # Upload the PDF file to s3
        self.s3.upload_file(str(self.path), self.BUCKET, self.path.name)

        # Start the document analysis
        r = self.textract.start_document_analysis(
            DocumentLocation={
                "S3Object": {"Bucket": self.BUCKET, "Name": self.path.name}
            },
            FeatureTypes=[
                "TABLES",
            ],
        )

        # Wait until job has finished
        jobstatus = None
        while jobstatus != "SUCCEEDED":
            response = self.textract.get_document_analysis(JobId=r["JobId"])
            jobstatus = response["JobStatus"]
            time.sleep(1)

        # Yield pg numbers and table
        yield from textract_to_words(response)



"""Abstract base class for performing ETL on PDF reports."""


from abc import ABC, abstractmethod
from pathlib import Path

import pandas as pd


class ETLPipeline(ABC):
    """
    An abstract base class to handle the extract-transform-load
    pipeline for parsing a PDF report.
    """

    @abstractmethod
    def extract(self) -> pd.DataFrame:
        """Extract the contents of the PDF page."""
        pass

    @abstractmethod
    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        """Transform the extracted data."""
        pass

    @abstractmethod
    def load(self, data: pd.DataFrame) -> None:
        """Load the transformed data into its data storage."""
        pass

    def _load_csv_data(self, data: pd.DataFrame, path: Path) -> None:
        """Internal function to load CSV data to a specified path."""

        if not path.parent.exists():
            path.parent.mkdir(parents=True)

        data.to_csv(path, index=False)

    @classmethod
    @abstractmethod
    def get_data_directory(cls, kind: str) -> str:
        """Internal function to get the file path.

        Parameters
        ----------
        kind : {'raw', 'processed'}
            type of data to load
        """
        pass

    def get_pdf_files(self):
        """Yield raw PDF file paths."""

        yield from sorted(self.get_data_directory("raw").glob("*.pdf"))

    def extract_transform(self) -> pd.DataFrame:
        """Convenience function to extract and then transform."""

        return self.transform(self.extract())

    def validate(self, data) -> bool:
        """
        Validate function to ensure data is correct before
        performing the "load" step.

        By default, returns True.
        """
        return True

    def extract_transform_load(self, validate: bool = True) -> pd.DataFrame:
        """
        Convenience function to extract, transform, and load.

        Optionally, validate the data before loading it.

        Parameters
        ----------
        validate :
            whether to run data validation prior to loading
        """
        # Extract & transform
        data = self.extract_transform()

        # Validate?
        if validate:
            assert self.validate(data) == True

        # Load the data
        self.load(data)

        return data



from .city import CityCollectionsReport
from .core import get_column_names


class CityOtherGovtsCollections(CityCollectionsReport):
    """Monthly City Other Governments Collections Report"""

    def transform(self, data):
        """Transform."""

        # Call base transform
        df = super().transform(data)

        # Get the start index
        start = df[0].str.contains("U.S. GOV", regex=True)
        start = start[start]
        assert len(start) == 1
        start = start.index[0]

        # Get the stop index
        stop = df[0].str.contains("TOTAL.*REVENUE.*GOV.*", regex=True)
        stop = stop[stop]
        assert len(stop) == 1
        stop = stop.index[0]

        # Trim
        df = df.loc[start:stop]

        # Check the length
        assert len(df) in [5, 7], f"length validation failed; length = {len(df)}"

        # Format the first column of names
        df[0] = (
            df[0]
            .str.strip()
            .str.lower()
            .str.replace("-", " ", regex=False)
            .str.replace(".", "", regex=False)
            .str.replace("&", "and", regex=False)
            .str.replace("\s+", "_", regex=True)
        )

        # Fix total
        sel = df[0].str.startswith("total_revenue_from")
        df.loc[sel, 0] = "total_revenue_other_govts"

        # Fix authorized adjustment
        sel = df[0].str.startswith("other_authorized")
        df.loc[sel, 0] = "other_authorized_adjustment"

        # Set the columns
        columns = get_column_names(self.month, self.year)
        columns = ["name"] + columns[-7:]
        df = df[[0] + list(df.columns[-7:])]
        df.columns = columns

        return df

    def load(self, data):
        """Load the data."""

        # Get the processed data path
        dirname = self.get_data_directory("processed")
        path = dirname / f"{self.year}-{self.month:02d}-other-govts.csv"

        # Load
        super()._load_csv_data(data, path)

    def validate(self, data):
        """Validate the input data."""

        # Trim to the month columns
        data = data.filter(regex=f"^{self.month_name}|name", axis=1)

        # Compare
        subcategories = data.query("name != 'total_revenue_other_govts'")
        total = data.query("name == 'total_revenue_other_govts'").squeeze()

        for col in data.columns:
            if col == "name":
                continue
            diff = subcategories[col].sum() - total[col]
            assert diff < 5

        return True



import calendar
import time
from contextlib import contextmanager
from pathlib import Path
from urllib.request import Request, urlopen

from bs4 import BeautifulSoup
from selenium import webdriver
from webdriver_manager.chrome import ChromeDriverManager

MONTH_LOOKUP = [x.lower() for x in calendar.month_abbr[1:]]


def parse_website(url):
    req = Request(url, headers={"User-Agent": "Mozilla/5.0"})
    web_byte = urlopen(req).read()
    return web_byte.decode("utf-8")


def extract_pdf_urls(url, id):

    # Parse the website
    soup = BeautifulSoup(parse_website(url), features="html.parser")

    # Get the id of the element
    rows = soup.select(f"table tr[id*={id}]")

    out = {}
    for row in rows:
        url = row.select_one("a")["href"]
        id = row["id"]
        month_name = id.split("-")[0][:3]
        month_num = MONTH_LOOKUP.index(month_name) + 1

        out[month_num] = url

    return out


def get_driver(dirname):

    options = webdriver.ChromeOptions()
    options.add_argument("--no-sandbox")
    options.add_argument("--headless")

    # Setup the download directory for PDFs
    profile = {
        "plugins.plugins_list": [
            {"enabled": False, "name": "Chrome PDF Viewer"}
        ],  # Disable Chrome's PDF Viewer
        "download.default_directory": dirname,
        "download.extensions_to_open": "applications/pdf",
    }
    options.add_experimental_option("prefs", profile)

    # Initialize with options
    driver = webdriver.Chrome(
        executable_path=ChromeDriverManager().install(), options=options
    )

    return driver


@contextmanager
def downloaded_pdf(driver, pdf_url, tmpdir, interval=1, time_limit=7):
    """Context manager to download a PDF to a local directory."""

    # Output path
    download_dir = Path(tmpdir)
    pdf_path = None

    try:
        # Get the PDF
        driver.get(pdf_url)

        # Initialize
        pdf_files = list(download_dir.glob("*.pdf"))
        total_sleep = 0
        while not len(pdf_files) and total_sleep <= time_limit:
            time.sleep(interval)
            total_sleep += interval
            pdf_files = list(download_dir.glob("*.pdf"))

        if len(pdf_files):
            pdf_path = pdf_files[0]
            yield pdf_path
        else:
            raise ValueError("PDF download failed")
    finally:

        # Remove the file after we are done!
        if pdf_path is not None and pdf_path.exists():
            pdf_path.unlink()



import calendar

import pandas as pd

from . import DATA_DIR
from .etl.collections import *
from .etl.utils.misc import fiscal_from_calendar_year


def load_sales_collections_by_sector() -> pd.DataFrame:
    """
    Load annual sales tax collections by sector

    Returns
    -------
    data :
        the annual sales collection data
    """
    # Get the path to the files to load
    dirname = SalesCollectionsByIndustry.get_data_directory("processed")
    files = dirname.glob("*.csv")

    out = []
    for f in files:

        # Get fiscal year
        fiscal_year = int(f"20{f.stem[2:]}")

        # Load the data
        df = (
            pd.read_csv(f)[["total", "industry", "parent_industry"]]
            .query("industry != 'Subtotal'")
            .assign(
                fiscal_year=fiscal_year,
            )
        )

        # Save
        out.append(df)

    # Combine multiple months
    out = pd.concat(out, ignore_index=True)

    return out.sort_values(
        ["fiscal_year", "parent_industry", "industry"], ascending=True
    )


def load_wage_collections_by_industry() -> pd.DataFrame:
    """
    Load monthly wage tax collections by industry

    Returns
    -------
    data :
        the monthly wage collection data
    """
    # Get the path to the files to load
    dirname = WageCollectionsByIndustry.get_data_directory("processed")
    files = dirname.glob("*.csv")

    out = []
    for f in files:

        # Get month/year
        year, month = map(int, f.stem.split("-"))
        month_name = calendar.month_abbr[month].lower()

        # Determine the fiscal year and tags
        fiscal_year = fiscal_from_calendar_year(month, year)
        this_FY = str(fiscal_year)[2:]
        last_FY = str(fiscal_year - 1)[2:]

        # Load the data and trim to "total"
        df = pd.read_csv(f)

        # Trim to the columns we want
        X = df[[f"{month_name}_{year}", "industry", "parent_industry"]]

        # Melt the data
        X = (
            X.melt(
                id_vars=["industry", "parent_industry"],
                value_name="total",
            )
            .assign(
                month_name=month_name,
                month=month,
                fiscal_month=((month - 7) % 12 + 1),
                year=year,
                fiscal_year=fiscal_year,
                total=lambda df: df.total.fillna(0),
            )
            .drop(labels=["variable"], axis=1)
        )

        # Save
        out.append(X)

    # Combine multiple months
    out = pd.concat(out, ignore_index=True)
    out["date"] = pd.to_datetime(
        out["month"].astype(str) + "/" + out["year"].astype(str)
    )

    return out.sort_values("date", ascending=True)


def _load_monthly_collections(files, total_only=False):
    """Internal function to load monthly collections data."""

    out = []
    for f in files:

        # Get month/year
        year, month, *_ = f.stem.split("-")
        year = int(year)
        month = int(month)
        month_name = calendar.month_abbr[month].lower()

        # Determine the fiscal year and tags
        fiscal_year = fiscal_from_calendar_year(month, year)
        this_FY = str(fiscal_year)[2:]
        last_FY = str(fiscal_year - 1)[2:]

        # Load the data
        df = pd.read_csv(f)
        if total_only:
            df = df.query("kind == 'total'")

        # Column names for this month
        a = f"{month_name}_fy{this_FY}"
        b = f"{month_name}_fy{last_FY}"

        # Trim to the columns we want
        X = df[["name", a, b]].rename(
            columns=dict(zip([a, b], [fiscal_year, fiscal_year - 1]))
        )

        # Melt the data
        X = (
            X.melt(id_vars=["name"], var_name="fiscal_year", value_name="total")
            .assign(
                month_name=month_name,
                month=month,
                fiscal_month=((month - 7) % 12 + 1),
                year=year,
            )
            .query(f"fiscal_year == {fiscal_year}")
        )
        X = X.fillna(0)

        # Save
        out.append(X)

    # Combine multiple months
    out = pd.concat(out, ignore_index=True)
    out["date"] = pd.to_datetime(
        out["month"].astype(str) + "/" + out["year"].astype(str)
    )

    return out.sort_values("date", ascending=True)


def load_monthly_other_govt_collections() -> pd.DataFrame:
    """
    Load monthly other govt collections for the City.

    Returns
    -------
    data :
        the monthly tax collection data
    """

    # Get the path to the files to load
    dirname = CityOtherGovtsCollections.get_data_directory("processed")
    files = dirname.glob("*-other-govts.csv")

    return _load_monthly_collections(files)


def load_monthly_nontax_collections() -> pd.DataFrame:
    """
    Load monthly nontax collections for the City.

    Returns
    -------
    data :
        the monthly tax collection data
    """

    # Get the path to the files to load
    dirname = CityNonTaxCollections.get_data_directory("processed")
    files = dirname.glob("*-nontax.csv")

    return _load_monthly_collections(files)


def load_monthly_tax_collections(kind: str) -> pd.DataFrame:
    """
    Load monthly tax collections, for either the City or School District.

    Parameters
    ----------
    kind :
        either 'city' or 'school'

    Returns
    -------
    data :
        the monthly tax collection data
    """

    assert kind in ["city", "school"]

    # Get the path to the files to load
    if kind == "city":
        dirname = CityTaxCollections.get_data_directory("processed")
    elif kind == "school":
        dirname = SchoolTaxCollections.get_data_directory("processed")
    files = dirname.glob("*-tax.csv")

    return _load_monthly_collections(files, total_only=True)



import tempfile
from urllib.error import HTTPError

import click
from loguru import logger

from .. import DATA_DIR
from ..collections import *
from .scrape import *


def _get_latest_raw_pdf(cls):
    """Given an ETL class, return the latest PDF in the data directory."""

    # Get PDF paths
    dirname = cls.get_data_directory("raw")
    pdf_files = dirname.glob("*.pdf")

    # Get the latest
    latest = sorted(pdf_files)[-1]
    year, month = map(int, latest.stem.split("_"))

    # Next month and year
    if month == 12:
        next_month = 1
        next_year = year + 1
    else:
        next_month = month + 1
        next_year = year

    return next_year, next_month


def _run_monthly_update(month, year, url, css_identifier, *etls):

    # Try to extract out the PDF links from the page
    try:
        pdf_urls = extract_pdf_urls(url, css_identifier)
    except HTTPError as err:
        if err.code == 404:
            logger.info(f"URL '{url}' does not exist")
            return None
        else:
            raise

    # Download and run ETL
    if month in pdf_urls:

        # Download to temp dir initially
        with tempfile.TemporaryDirectory() as tmpdir:

            # Get the driver
            driver = get_driver(tmpdir)

            # The remote URL
            remote_pdf_path = pdf_urls[month]

            # Log
            logger.info(f"Downloading PDF from '{remote_pdf_path}'")

            # Local path
            dirname = etls[0].get_data_directory("raw")
            local_pdf_path = dirname / f"{year}_{month:02d}.pdf"

            # Download the PDF
            with downloaded_pdf(
                driver, remote_pdf_path, tmpdir, interval=1
            ) as pdf_path:

                if not local_pdf_path.parent.exists():
                    local_pdf_path.parent.mkdir()

                pdf_path.rename(local_pdf_path)

            # Run the ETL
            try:
                for cls in etls:

                    # Log
                    logger.info(f"Running ETL for {cls.__name__}")

                    # Run the ETL
                    report = cls(year=year, month=month)
                    report.extract_transform_load()
            except Exception:

                if local_pdf_path.exists():
                    local_pdf_path.unlink()
                raise

    else:
        logger.info(f"...no updates found")


@click.command()
@click.version_option()
def update_monthly_wage_collections():
    """Check for updates to the monthly wage collection report."""

    # Get the month/year of next PDF to look for
    year, month = _get_latest_raw_pdf(WageCollectionsByIndustry)

    # Log
    logger.info(f"Checking for PDF report for month '{month}' and year '{year}'")

    # Extract out PDF urls on the city's website
    url = f"https://www.phila.gov/documents/{year}-wage-tax-by-industry/"
    css_identifier = "wage-taxes"

    # Run the update
    _run_monthly_update(month, year, url, css_identifier, WageCollectionsByIndustry)


@click.command()
@click.version_option()
def update_monthly_city_collections():
    """Check for updates to the monthly city collection report."""

    # Get the month/year of next PDF to look for
    year, month = _get_latest_raw_pdf(CityTaxCollections)

    # Get the fiscal year
    if month < 7:
        fiscal_year = year
    else:
        fiscal_year = year + 1

    # Log
    logger.info(f"Checking for PDF report for month '{month}' and year '{year}'")

    # Extract out PDF urls on the city's website
    url = f"https://www.phila.gov/documents/fy-{fiscal_year}-city-monthly-revenue-collections/"
    css_identifier = "revenue-collections"

    # Run the update
    _run_monthly_update(
        month,
        year,
        url,
        css_identifier,
        CityTaxCollections,
        CityNonTaxCollections,
        CityOtherGovtsCollections,
    )


@click.command()
@click.version_option()
def update_monthly_school_collections():
    """Check for updates to the monthly school district collection report."""

    # Get the month/year of next PDF to look for
    year, month = _get_latest_raw_pdf(SchoolTaxCollections)

    # Get the fiscal year
    if month < 7:
        fiscal_year = year
    else:
        fiscal_year = year + 1

    # Log
    logger.info(f"Checking for PDF report for month '{month}' and year '{year}'")

    # Extract out PDF urls on the city's website
    url = f"https://www.phila.gov/documents/fy-{fiscal_year}-school-district-monthly-revenue-collections/"
    css_identifier = "revenue-collections"

    # Run the update
    _run_monthly_update(month, year, url, css_identifier, SchoolTaxCollections)



from typing import List, Optional

import numpy as np
import pandas as pd


def convert_to_floats(
    df: pd.DataFrame, usecols: Optional[List[str]] = None, errors: str = "coerce"
) -> pd.DataFrame:
    """
    Convert string values in currency format to floats.

    Parameters
    ----------
    df :
        the data to format
    usecols :
        an optional list of columns to convert
    errors : {‘ignore’, ‘raise’, ‘coerce’}, default ‘raise’
        - If ‘raise’, then invalid parsing will raise an exception.
        - If ‘coerce’, then invalid parsing will be set as NaN.
        - If ‘ignore’, then invalid parsing will return the input.
    """
    if usecols is None:
        usecols = df.columns

    for col in usecols:
        df[col] = pd.to_numeric(
            df[col].replace("[\$,)]", "", regex=True).replace("[(]", "-", regex=True),
            errors=errors,
        )
    return df


def remove_footnotes(df: pd.DataFrame) -> pd.DataFrame:
    """Remove any lines starting with an asterisk."""
    is_footnote = df[0].str.strip().str.startswith("*")
    return df.loc[~is_footnote].copy()


def replace_missing_cells(df: pd.DataFrame) -> pd.DataFrame:
    """Fill empty or N/A values with NaN values."""

    df = df.replace("", np.nan)
    for col in df.columns[1:]:
        df[col] = df[col].replace("N/A", np.nan)
        df[col] = df[col].replace("-", np.nan)

    return df


def fix_percentages(df: pd.DataFrame) -> pd.DataFrame:
    """Strip percent signs from the end of cell values."""
    for col in df.columns[1:]:
        df[col] = df[col].str.strip().str.rstrip("%")
    return df


def strip_dollar_signs(df: pd.DataFrame) -> pd.DataFrame:
    """Strip dollars signs from the start of cell values."""
    for col in df.columns[1:]:
        df[col] = df[col].str.strip().str.lstrip("$")
    return df


def remove_spaces(df: pd.DataFrame) -> pd.DataFrame:
    """Strip out spaces from cells."""
    for col in df.columns[1:]:
        df[col] = df[col].str.replace("\s+", "", regex=True)
    return df


def fix_duplicate_parens(df: pd.DataFrame) -> pd.DataFrame:
    """Change any duplicate parentheses to single only."""
    for col in df.columns[1:]:
        df[col] = df[col].str.replace("((", "(", regex=False)
        df[col] = df[col].str.replace("))", ")", regex=False)
    return df


def remove_missing_rows(df: pd.DataFrame, usecols=None) -> pd.DataFrame:
    """Remove rows that are empty."""
    if usecols is None:
        usecols = df.columns
    missing = df[usecols].isnull().all(axis=1)
    return df.loc[~missing]


def remove_extra_letters(df: pd.DataFrame, col_num: int) -> pd.DataFrame:
    """Remove any letters mixed into the first column."""
    col = df.columns[col_num]
    df[col] = df[col].str.replace("[A-Z]", "", regex=True)
    return df


def fix_duplicated_chars(data: pd.DataFrame) -> pd.DataFrame:
    """Check for duplicated characters."""

    # Check for rows where characters are duplicated
    first = data[1][data[1].str.len() > 0]
    first = first.str.replace(r"([a-zA-Z0-9%,.])\1+", r"", regex=True).str.replace(
        ",", ""
    )

    # Use regex to remove any cells that have characters duplicated
    duplicated = first.loc[first.str.len() == 0].index
    for index in duplicated:
        data.loc[index] = data.loc[index].str.replace(
            r"([a-zA-Z0-9%,.])\1{1,2}", r"\1", regex=True
        )

    return data


def remove_empty_columns(data, use_nan=False):
    """Remove empty columns."""

    # Remove null columns
    if use_nan:
        bad_columns = data.isnull().all()
    else:
        bad_columns = (data == "").all()

    if bad_columns.sum():
        data = data[data.columns[~bad_columns]]
        data.columns = list(range(0, len(data.columns)))

    return data



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['phl_budget_data',
 'phl_budget_data.etl',
 'phl_budget_data.etl.collections',
 'phl_budget_data.etl.collections.by_industry',
 'phl_budget_data.etl.collections.monthly',
 'phl_budget_data.etl.sentinel',
 'phl_budget_data.etl.utils']

package_data = \
{'': ['*'],
 'phl_budget_data': ['data/processed/collections/by-industry/wage/*',
                     'data/processed/collections/monthly/city/*',
                     'data/processed/collections/monthly/school/*']}

install_requires = \
['beautifulsoup4>=4.9.3,<5.0.0',
 'boto3>=1.17.12,<2.0.0',
 'click>=7.0,<8.0',
 'desert>=2020.11.18,<2021.0.0',
 'intervaltree>=3.1.0,<4.0.0',
 'loguru>=0.5.3,<0.6.0',
 'numpy==1.18.0',
 'pandas>=1.2.1,<2.0.0',
 'pdfplumber>=0.5.25,<0.6.0',
 'python-dotenv>=0.15.0,<0.16.0',
 'selenium>=3.141.0,<4.0.0',
 'webdriver-manager>=3.3.0,<4.0.0']

entry_points = \
{'console_scripts': ['phl-monthly-collections-etl = '
                     'phl_budget_data.__main__:monthly_collections_etl',
                     'phl-update-monthly-city-collections = '
                     'phl_budget_data.etl.sentinel.core:update_monthly_city_collections',
                     'phl-update-monthly-school-collections = '
                     'phl_budget_data.etl.sentinel.core:update_monthly_school_collections',
                     'phl-update-monthly-wage-collections = '
                     'phl_budget_data.etl.sentinel.core:update_monthly_wage_collections']}

setup_kwargs = {
    'name': 'phl-budget-data',
    'version': '0.1.1',
    'description': 'PHL Budget Data',
    'long_description': '# phl-budget-data\n\nAggregating and cleaning City of Philadelphia budget-related data\n\n## Installation\n\n```\npip install phl_budget_data\n```\n## Examples\n\n',
    'author': 'Nick Hand',
    'author_email': 'nick.hand@phila.gov',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/PhiladelphiaController/phl-budget-data',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



import numpy as np
import pandas as pd
import pdfplumber

from ...utils.pdf import extract_words, words_to_table
from ...utils.transformations import remove_empty_columns
from .core import MonthlyCollectionsReport


def find_top_cutoff(pg):
    """Search for the top cutoff of header on the page."""

    # Extract text to look for a header
    pg_text = pg.extract_text()

    # This shows up in the header columns
    top = 0
    has_header = "FISCAL YEAR TO DATE" in pg_text
    if has_header:

        # Determine if header has 1 or 2 horizontal lines
        # We want to find the bottom horizontal line to trim
        rects = sorted(pg.rects, key=lambda x: x["y0"], reverse=True)
        if "Comparative Statement" in pg_text:
            top = rects[1]["bottom"]
        else:
            top = rects[0]["bottom"]

    return top


def find_footer_cutoff(pg):
    """Search for a horizontal line separating footer from data."""

    rects = [r for r in pg.rects if r["width"] / pg.width > 0.5]
    max_rect = max(rects, key=lambda r: r["bottom"])
    if max_rect["bottom"] > 0.9 * float(pg.height):
        return max_rect
    else:
        return None


class CityCollectionsReport(MonthlyCollectionsReport):
    """
    Monthly City Collections Report.

    Parameters
    ----------
    month :
        the calendar month number (starting at 1)
    year :
        the calendar year
    """

    report_type = "city"

    @property
    def legacy(self):
        """Whether the format is the legacy or current version."""
        return self.num_pages > 4

    def extract(self) -> pd.DataFrame:
        """Parse and extract the contents of a legacy PDF page."""

        # Open the PDF document
        with pdfplumber.open(self.path) as pdf:

            # Loop over each page
            out = []
            for pg in pdf.pages:

                # Is there a width-spanning line at the top?
                top = find_top_cutoff(pg)

                # Is there a width-spanning line at the bottom?
                footer_cutoff = find_footer_cutoff(pg)
                if footer_cutoff is not None:
                    bottom = footer_cutoff["bottom"]
                else:
                    bottom = pg.height

                # Crop the main part of the document and extract the words
                cropped = pg.crop([0, top, pg.width, bottom])
                words = extract_words(
                    cropped, keep_blank_chars=False, x_tolerance=1, y_tolerance=1
                )

                # Group the words into a table
                data = words_to_table(
                    words,
                    text_tolerance_y=5,
                    text_tolerance_x=5,
                    column_tolerance=20,
                    min_col_sep=24,
                    header_column_overlap=20,
                )

                # Remove first row of header if we need to
                for phrase in ["prelim", "final", "budget"]:
                    sel = data[0].str.lower().str.startswith(phrase)
                    data = data.loc[~sel]

                # Remove empty columns
                data = remove_empty_columns(data, use_nan=False)

                # Check number of columns
                if len(out):
                    if len(data.columns) != len(out[-1].columns):
                        raise ValueError("Column mismatch when parsing multiple pages")

                # Save it
                out.append(data)

            # Return concatenation
            return pd.concat(out, axis=0, ignore_index=True)
